import json
import dateutil.parser
import numpy as np

from frcm.weatherdata.extractor import Extractor
from frcm.datamodel.model import *


class METExtractor(Extractor):

    def extract_observations(self, frost_response_str: str, location: Location) -> Observations:

        frost_response = json.loads(frost_response_str)
        data_list = frost_response['data']

        weatherdatapoints = list()

        source_id = None

        if len(data_list) > 1:

            source_id = data_list[0]['sourceId']

            for data in data_list:

                reference_time = dateutil.parser.parse(data['referenceTime'])
                station_observations = data['observations']

                temperature = np.nan
                relative_humidity = np.nan
                wind_speed = np.nan

                for station_observation in station_observations:

                    # string to datatime object required
                    timestamp = reference_time # assume that observations have the same time stamp

                    # TODO: rewrite to use a switch
                    if station_observation['elementId'] == 'air_temperature':
                        temperature = station_observation['value']
                    elif station_observation['elementId'] == 'relative_humidity':
                        relative_humidity = station_observation['value']
                    elif station_observation['elementId'] == 'wind_speed':
                        wind_speed = station_observation['value']

                wd_point = WeatherDataPoint(temperature=temperature,
                                            humidity=relative_humidity,
                                            wind_speed=wind_speed,
                                            timestamp=timestamp
                                            )

                weatherdatapoints.append(wd_point)

        # TODO: maybe also source as part of the parameters - or extract weather data function instead
        observations = Observations(source=source_id, location=location,data=weatherdatapoints)

        return observations

    def extract_forecast(self, met_response_str: str) -> Forecast:

        met_response = json.loads(met_response_str)

        coordinates = met_response['geometry']['coordinates']

        latitude = coordinates[1]
        longitude = coordinates[0]

        location = Location(latitude=latitude, longitude=longitude)

        timeseries = met_response['properties']['timeseries']

        weatherdatapoints = list()

        for forecast in timeseries:

            timestamp = dateutil.parser.parse(forecast['time'])

            details = forecast['data']['instant']['details']

            temperature = details['air_temperature']
            humidity = details['relative_humidity']
            wind_speed = details['wind_speed']

            wd_point = WeatherDataPoint(temperature=temperature,
                                        humidity=humidity,
                                        wind_speed=wind_speed,
                                        timestamp=timestamp)

            weatherdatapoints.append(wd_point)

        forecast = Forecast(location=location,data=weatherdatapoints)

        return forecast

    def extract_weatherdata(self, frost_response: str, met_response: str, location: Location):

        observations = self.extract_observations(frost_response, location)

        forecast = self.extract_forecast(met_response)

        now = datetime.datetime.now()  # FIXME: date from each response should be used

        weather_data = WeatherData(created=now,
                                   observations=observations,
                                   forecast=forecast)

        return weather_data



import datetime

from pydantic import BaseModel


class Location(BaseModel):

    latitude: float
    longitude: float


class WeatherDataPoint(BaseModel):

    temperature: float
    humidity: float
    wind_speed: float
    timestamp: datetime.datetime

    def __str__(self):

        format_str = f'WeatherData[{self.timestamp}] {self.temperature, self.humidity, self.wind_speed}]'

        return format_str


class Observations(BaseModel):

    source: str
    location: Location
    data: list[WeatherDataPoint]

    def __str__(self):
        format_str = f'Observations [Source: {self.source} @ Location: {self.location}]\n'

        # Join all data points using '\n' as a separator
        data_strings = '\n'.join(map(str, self.data))

        return format_str + data_strings + '\n'


class Forecast(BaseModel):

    location: Location
    data: list[WeatherDataPoint]

    def __str__(self):
        format_str = f'Forecast @ Location: {self.location}\n'

        # Join all data points using '\n' as a separator
        data_strings = '\n'.join(map(str, self.data))

        return format_str + data_strings + '\n'


class WeatherData(BaseModel):

    created: datetime.datetime

    observations: Observations
    forecast: Forecast

    def to_json(self):
        return self.model_dump_json()


class FireRisk(BaseModel):

    timestamp: datetime.datetime
    ttf: float
    wind_speed: float

    def __str__(self):
        format_str = f'FireRisks[{self.timestamp} TTF({self.ttf}) WindSpeed({self.wind_speed})]'

        return format_str


class FireRiskPrediction(BaseModel):

    location: Location
    firerisks: list[FireRisk]

    def __str__(self):
        format_str = f'FireRiskPrediction[{self.location}]\n'

        # Generate list of formatted data points
        data_str = [str(data_point) for data_point in self.firerisks]

        # Join the list of formatted data points
        format_str += '\n'.join(data_str)

        return format_str



from frcm.datamodel.model import *
from frcm.fireriskmodel.parameters import delta_t
import numpy as np


def combine_obs_fct(sorted_data_obs, sorted_data_fct, parameter):
    obs = [getattr(obj, parameter) for obj in sorted_data_obs]
    fct = [getattr(obj, parameter) for obj in sorted_data_fct]
    combined = obs + fct
    return combined


def clean_nan(data_vector, time_vector):
    # Treat Data
    nan_indices = np.where(np.isnan(data_vector))
    data_cleaned = np.delete(data_vector, nan_indices)
    # Treat associated time_vector
    time_cleaned = np.delete(time_vector, nan_indices)
    return data_cleaned, time_cleaned


def find_data_gap(*args): # np.array
    max_delta = []
    for arr in args:
        differences = np.diff(arr)
        max_delta.append(np.max(differences))
    return np.max(max_delta)


def preprocess(wd: WeatherData):

    # Should not be necessary, but data is initially sorted according to the timestamps
    sorted_data_obs = sorted(wd.observations.data, key=lambda x: x.timestamp)
    sorted_data_fct = sorted(wd.forecast.data, key=lambda x: x.timestamp)

    # Combine data
    timestamp_vector = combine_obs_fct(sorted_data_obs, sorted_data_fct, 'timestamp')
    temp_vector = combine_obs_fct(sorted_data_obs, sorted_data_fct, 'temperature')
    humidity_vector = combine_obs_fct(sorted_data_obs, sorted_data_fct, 'humidity')
    wind_vector = combine_obs_fct(sorted_data_obs, sorted_data_fct, 'wind_speed')

    # Convert timestamp vector to a vector containing delta time of adjacent elements in seconds, starting at 0.
    timestamp_vector_sec = [round((timestamp - timestamp_vector[0]).total_seconds()) for timestamp in timestamp_vector]
    # Get start of computation as datetime
    start_time = timestamp_vector[0]

    # Identify position of np.nan values and remove from "parameter"_vector and associated "parameter"_timevector
    # Resulting vectors are used in the np interpolation function (np.interp)
    temp_clean, time_temp_clean = clean_nan(data_vector=temp_vector, time_vector=timestamp_vector_sec)
    humidity_clean, time_humidity_clean = clean_nan(data_vector=humidity_vector, time_vector=timestamp_vector_sec)
    wind_clean, time_wind_clean = clean_nan(data_vector=wind_vector, time_vector=timestamp_vector_sec)

    # Create interpolation time vector in seconds. This vector contains all the datapoints for which the np.interp-
    # function shall provide interpolated values.
    interpolation_timevector_sec = [i for i in range(timestamp_vector_sec[0], timestamp_vector_sec[-1] + 1, delta_t)]

    # Find largest gap in data. Currently only considering temperature and humidity. Delta is given in seconds.
    max_time_delta = find_data_gap(time_temp_clean, time_humidity_clean)

    # Interpolate all data
    temp_interpolated = np.interp(interpolation_timevector_sec, time_temp_clean, temp_clean)
    humidity_interpolated = np.interp(interpolation_timevector_sec, time_humidity_clean, humidity_clean)
    wind_interpolated = np.interp(interpolation_timevector_sec, time_wind_clean, wind_clean)

    return start_time, interpolation_timevector_sec, temp_interpolated, humidity_interpolated, wind_interpolated, max_time_delta




import math

import frcm.fireriskmodel.parameters as mp

"""Functions for computing saturation vapor pressure, and water concentrations"""
""" pw_sat -- cw_sat -- cw_in -- initial fmc """


# saturation vapor pressure at temperature Temp_c (celsius)
def calc_pwsat(temp_c) -> float:
    pwsat = 610.78 * math.exp((17.2694 * temp_c) / (temp_c + 238.3))
    return pwsat


# saturation water concentration based on saturated vapor pressure (pwsat)
def calc_cwsat(pwsat,temp_c) -> float:
    cwsat = (pwsat * mp.mol_weight)/(mp.gas_constant * (temp_c + 273.15))
    return cwsat


# actual water concentration in air
def calc_cw(rh, cwsat):
    cw = rh / 100 * cwsat
    return cw


# computes initial fmc of wooden panels
# computes fmc from indoor rh (equilibrium state) rh must be given as a fraction, e.g., 0.35
def calc_fmc(rh) -> float:
    c_fmc = 0.0017 + 0.2524 * rh - 0.1986 * math.pow(rh,2) + 0.0279 * math.pow(rh,3) + 0.167 * math.pow(rh,4)
    return c_fmc


""" Functions related to ventilation """
""" Air change per Hour -- Beta"""


# air change per hour (ach)
def calc_ach(temp_c_out, temp_c_in) -> float:
    c_ach = mp.gamma * math.sqrt((abs(1 / (temp_c_out + 273.15) - 1 / (temp_c_in + 273.15)) / (temp_c_out + 273.15)))
    return c_ach


# beta ventilation factor
def calc_beta(c_ach) -> float:
    c_beta = 1 - math.exp((-c_ach * mp.delta_t)/3600)
    return c_beta


""" Updating variables post the modelling of wooden panel humidity transport (wall modelling) """


# extrapolating wooden panel fmc to a surface value
def calc_surf(c1_t, c2_t) -> float:
    c_surf = c1_t - 0.5 * (c2_t - c1_t)
    return c_surf


# water concentration difference between bulk air and wall boundary layer - inputs from previous timestep
def calc_deltac(rhin, rhwall, cwsatin) -> float:
    deltac = (rhwall - rhin) * cwsatin
    return deltac


# relative humidity at wooden panel surfaces - inputs surface fmc at equal timestep
def calc_rhwall(cfmc) -> float:
    rhwall = 0.0698 - 1.258 * (cfmc / mp.rho_wood) + 125.35 * math.pow((cfmc / mp.rho_wood),2) - 809.43 * math.pow((cfmc / mp.rho_wood),3) + 1583.8 * math.pow((cfmc / mp.rho_wood),4)
    return rhwall


# indoor water concentration - input from previous timestep
def calc_cwin(cac, cwall, csupply, cwin, beta):
    cwin = (1 - beta) * cwin + cac + cwall + csupply
    return cwin


""" Calculation of wooden panel humidity transport """


# computing the fmc in layer 1 (c1_t+1) (facing interior of enclosure) by use of fmc values in layer 1 and layer 2
# inputs from previous timestep
def calc_layer1(rhin,rhwall,c1_t,c2_t, csatin) -> float:
    layer1 = c1_t + (mp.delta_t/mp.delta_x) * ((mp.D_W_a / mp.boundary_layer) * (rhin - rhwall) * csatin + (mp.D_w_s / mp.delta_x) * (c2_t - c1_t))
    return layer1


# computing the fmc in layers 2 to N-1 (second last layer) at time t+1 by second order central difference - inputs from
# previous timestep
def calc_middle_layers(cn_t, c_prev_n_t, c_post_n_t) -> float:
    middle_layer = cn_t + mp.fourier * (c_prev_n_t - 2*cn_t + c_post_n_t)
    return middle_layer


# computing the fmc in the last layer (backside of wooden panels) based on fmc from layer N-1 and Layer N both from
# previous timestep
def calc_outer_layer(cn_t, c_pre_n_t) -> float:
    outer_layer = cn_t + mp.fourier * (c_pre_n_t - cn_t)
    return outer_layer


""" The main contributions to change in indoor water concentration """
""" Supply -- Air Change by Ventilation -- Humidity Exchange From Wooden Surfaces"""


def calc_csupply(sup) -> float:
    csupply = sup / mp.Vol
    return csupply


def calc_cac(beta, cw_out, temp_c_out, temp_c_in) -> float:
    cac = beta * cw_out * ((temp_c_out + 273.15) / (temp_c_in + 273.15))
    return cac


def calc_cwall(deltac) -> float:
    cwall = (mp.A_ex * mp.D_W_a * deltac * mp.delta_t / mp.boundary_layer) / mp.Vol
    return cwall



import datetime
import dateutil.parser

import frcm.datamodel.model as dm


def min_time(data: list[dm.WeatherDataPoint]):
    # assume that data is time sorted in ascending order and non-empty
    return data[0].timestamp


def max_time(data: list[dm.WeatherDataPoint]):
    # assume that data is time sorted in ascending order and non-empty
    return data[-1].timestamp


def is_sorted(data: list[dm.WeatherDataPoint]) -> bool:
    i = 1
    ascending = True

    while i < len(data) and ascending:
        ascending = data[i].timestamp > data[i - 1].timestamp

        i = i + 1

    return ascending


def within_timedelta(data: list[dm.WeatherDataPoint], max_delta: datetime.timedelta) -> bool:
    i = 1
    violated = False

    while i < len(data) and (not violated):
        violated = (data[i].timestamp - data[i - 1].timestamp) < max_delta

        i = i + 1

    return violated


def dict_to_wdp(wdp) -> dm.WeatherDataPoint:

    wdp = dm.WeatherDataPoint(
        temperature=wdp['temperature'],
        humidity=wdp['humidity'],
        wind_speed=wdp['wind_speed'],
        timestamp=dateutil.parser.parse(wdp['timestamp']))

    return wdp


def list_to_wdps(wdps) -> list[dm.WeatherDataPoint]:

    return list(map(dict_to_wdp, wdps))


def wdps_list_str(wdps : list[dm.WeatherDataPoint]) -> str:
    # TODO: current string concatenation is inefficient
    format_str = ''
    for wdp in wdps:
        format_str = format_str + str(wdp) + '\n'

    return format_str


# TODO: validate could possibly raise different exceptions depending on violation
# TODO: check also that created is between max_time and min_time
# TODO: check that the required data is also present in the data points

def wd_validate(wd: dm.WeatherData, max_delta: datetime.timedelta):

    max_time_obs = max_time(wd.observations.data)
    min_time_fct = min_time(wd.forecast.data)

    is_sorted_obs = is_sorted(wd.observations.data)
    is_sorted_fct = is_sorted(wd.forecast.data)

    is_delta_obs = within_timedelta(wd.observations.data, max_delta)
    is_delta_fct = within_timedelta(wd.forecast.data, max_delta)

    return ((max_time_obs < min_time_fct) and
            is_sorted_obs and is_sorted_fct and
            is_delta_obs and is_delta_fct and
            (max_time_obs - min_time_fct) < max_delta)



import dateutil.parser

from frcm.datamodel.model import WeatherDataPoint


def weatherdata_parse(datadict) -> list[WeatherDataPoint]:

    data = list()

    for item in datadict:

        temperature = item['temperature']
        humidity = item['humidity']
        wind_speed = item['wind_speed']
        timestamp = dateutil.parser.parse(item['timestamp'])

        wd_point = WeatherDataPoint(temperature=temperature,
                                    humidity=humidity,
                                    wind_speed=wind_speed,
                                    timestamp=timestamp)

        data.append(wd_point)

    return data




import numpy as np
import datetime

import frcm.datamodel.model as dm
import frcm.fireriskmodel.parameters as mp
import frcm.fireriskmodel.utils as func
import frcm.fireriskmodel.preprocess as pp


def compute(wd: dm.WeatherData) -> dm.FireRiskPrediction:

    # Get interpolated values #TODO (NOTE) The max_time_delta represents the largest gap in missing data (seconds). It can be used to provide suited warning/error message.
    start_time, time_interpolated_sec, temp_interpolated, humidity_interpolated, wind_interpolated, max_time_delta = pp.preprocess(wd)
    comp_loc = wd.forecast.location

    # Compute RH_in and TTF
    rh_in, ttf = compute_fr(temp_interpolated, humidity_interpolated)

    # Reduce data to once per hour, but the time is still given as seconds
    rf = int(
        3600 / mp.delta_t)  # Reduction factor, i.e., how many intervals per hour. Default delta_t = 720 s, hence rf = 5.
    rh_in_hour = rh_in[::rf]  # Average is not computed, values are extracted per hour.
    ttf_in_hour = ttf[::rf]
    time_in_hour = time_interpolated_sec[::rf] # Time is still in seconds but given for every hour.
    wind_speed_in_hour = wind_interpolated[::rf]

    # Create response according to datamodel
    firerisks = []
    for i in range(len(rh_in_hour)):
        timestamps = start_time + datetime.timedelta(seconds=time_in_hour[i])
        firerisk_i = dm.FireRisk(timestamp=timestamps, ttf=ttf_in_hour[i], wind_speed=wind_speed_in_hour[i])
        firerisks.append(firerisk_i)

    FireRiskResponse = dm.FireRiskPrediction(location=comp_loc, firerisks=firerisks)

    return FireRiskResponse


def compute_fr(temp_c_out, rh_out):

    "Indoor temperature vector"
    temp_c_in = [mp.T_c_in] * len(temp_c_out)  # Potential future changes may involve dynamic in-home temperatures

    """ compute saturation vapor pressures and water concentrations, outdoor and indoor """
    # w = water, sat = saturation
    pw_sat_out = list(map(func.calc_pwsat, temp_c_out))
    cw_sat_out = list(map(func.calc_cwsat, pw_sat_out, temp_c_out))
    cw_out = list(map(func.calc_cw, rh_out, cw_sat_out))
    pw_sat_in = list(map(func.calc_pwsat, temp_c_in))
    cw_sat_in = list(map(func.calc_cwsat, pw_sat_in, temp_c_in))

    # ventilation variables
    ach = list(map(func.calc_ach, temp_c_out, temp_c_in))
    beta = list(map(func.calc_beta, ach))

    # calculate supply per timestep and make a supply vector (for future updates - currently containing constant values)
    supply_pts = (mp.supply_24h / (24 * 3600)) * mp.delta_t
    supply = [supply_pts] * len(temp_c_out)

    """ create wall array and vector """

    # wall array - storing fmc values - rows represent a step in time, columns represent wooden panel layers.
    wall = np.zeros(shape=(len(temp_c_out), mp.sub_layers))
    # used to update the wall array
    wall_vector = np.zeros(mp.sub_layers)

    # initial fmc value in wooden panels
    initial_fmc = func.calc_fmc(mp.RH_in) * mp.rho_wood

    """ placeholders """
    # shall contain wooden surface fmc values
    surface = np.zeros(len(temp_c_out))
    # shall contain wooden surface (boundary layer) rh values
    rh_wall = np.zeros(len(temp_c_out))
    # shall contain bulk air (in-home) rh values
    rh_in = np.zeros(len(temp_c_out))
    # shall contain bulk air (in-home) water concentrations
    cw_in = np.zeros(len(temp_c_out))
    # shall contain water concentration difference between rh_wall and rh_in
    delta_c = np.zeros(len(temp_c_out))
    # shall contain water concentration (contribution) per timestep from wooden surfaces
    c_wall = np.zeros(len(temp_c_out))


    # set initial conditions
    wall[0] = [initial_fmc] * mp.sub_layers
    surface[0] = func.calc_surf(wall[0][0], wall[0][1])
    rh_wall[0] = func.calc_rhwall(surface[0])
    rh_in[0] = mp.RH_in
    cw_in[0] = mp.RH_in * cw_sat_in[0]
    delta_c[0] = func.calc_deltac(rh_in[0], rh_wall[0], cw_sat_in[0])
    c_wall[0] = func.calc_cwall(delta_c[0])

    c_ac = list(map(func.calc_cac, beta, cw_out, temp_c_out, temp_c_in))
    c_supply = (list(map(func.calc_csupply, supply)))

    for i in range(len(temp_c_out) - 1):
        # compute fmc in layer 1
        wall_vector[0] = func.calc_layer1(rh_in[i], rh_wall[i], wall[i][0], wall[i][1], cw_sat_in[i])
        n = 0
        for l in range(mp.sub_layers-2):
            # compute fmc in wall layers 2 to N-1
            n = n + 1
            wall_vector[n] = func.calc_middle_layers(wall[i][n], wall[i][n - 1], wall[i][n + 1])
        # compute fmc in wall layer N (panel backside)
        wall_vector[-1] = func.calc_outer_layer(wall[i][-1], wall[i][-2])
        # update wall array
        wall[i + 1][:] = wall_vector
        # update surface vector
        surface[i + 1] = func.calc_surf(wall[i + 1][0], wall[i + 1][1])
        # update rh_wall
        rh_wall[i + 1] = func.calc_rhwall(surface[i + 1])
        # update water concentration difference between bulk air and boundary layer
        delta_c[i + 1] = func.calc_deltac(rh_in[i], rh_wall[i], cw_sat_in[i])
        # update indoor water concentration
        cw_in[i + 1] = func.calc_cwin(c_ac[i], c_wall[i], c_supply[i], cw_in[i], beta[i])
        # update indoor relative humidity
        rh_in[i + 1] = cw_in[i + 1] / cw_sat_in[i + 1]
        # update c_wall
        c_wall[i + 1] = func.calc_cwall(delta_c[i + 1])

    # Compute ttf
    factor = 100 / mp.rho_wood
    fmc = list(map(lambda x: x*factor, surface))
    ttf = list(map(lambda y: 2 * np.exp(0.16*y),fmc))

    return rh_in, ttf





import abc

from frcm.datamodel.model import Observations, Forecast


class Extractor:
    @abc.abstractmethod
    def extract_observations(self, data: str) -> Observations:
        pass

    @abc.abstractmethod
    def extract_forecast(self, data: str) -> Forecast:
        pass






# General modelling parameters

panel_thickness = 0.012                 # m - panel thickness
sub_layers = 10                         # number of sub layers
delta_x = panel_thickness / sub_layers  # m - thickness of modelled sublayer in wooden panels
D_w_s = 3.0 * 10**(-10)                 # m^2/s - diffusion coefficient of water in wood
D_W_a = 2.5 * 10**(-5)                 # m^2/s - diffusion coefficient of water in air at 22'C
delta_t = 720                           # seconds - modelling timestep (constant)
boundary_layer = 0.01                   # m - boundary layer at wooden wall surface (constant)
gas_constant = 8.314                    # J/(Kg * K) Universal gas constant
mol_weight = 0.018015                   # Kg/mol - molecular weight of water vapor (constant)
fourier = 0.15                          # Fourier number




# Model specific parameters (generic wooden home enclosure describing a combined living room and kitchen)

T_k_in = 295.15         # K - indoor temperature (constant)
T_c_in = 22             # 'C - indoor temperature (constant)
RH_in = 0.35            # initial estimate/guess (variable)
A_ex = 60               # m^2 - wooden area exchanging humidity with the enclosure (constant)
Vol = 120               # m^2 - volume of enclosure (constant)
A_V_Ratio = A_ex/Vol    # ratio of wooden panel exchange area to enclosure volume (constant)
supply_24h = 1          # Kg/24 hour - supplied water vapor from cooking, respiring, plants++
rho_wood = 500          # kg/m^3 - density of wood (constant)
gamma = 380             # Ventilation constant




import abc

from frcm.datamodel.model import *


class WeatherDataClient:

    # TODO: add variants for time period on observations and timedelta on forecast

    @abc.abstractmethod
    def fetch_observations(self, location: Location) -> Observations:
        pass

    @abc.abstractmethod
    def fetch_forecast(self, location: Location) -> Forecast:
        pass



import datetime

from frcm.datamodel.model import FireRiskPrediction, Location, WeatherData, Observations, Forecast
from frcm.weatherdata.client import WeatherDataClient
import frcm.fireriskmodel.compute

from frcm.weatherdata.client_met import METClient
from frcm.weatherdata.extractor_met import METExtractor


class FireRiskAPI:

    def __init__(self, client: WeatherDataClient):
        self.client = client
        self.timedelta_ok = datetime.timedelta(days=1) # TODO: when during a day is observations updated? (12:00 and 06:00)
        # TODO (NOTE): Short term forecast updates every 3rd hour with long term forecast every 12th hour at 12:00 and 06:00
        self.interpolate_distance = 720

    def compute(self, wd: WeatherData) -> FireRiskPrediction:

        return frcm.fireriskmodel.compute.compute(wd)

    def get_wd_observations_to_now(self, location: Location, time_now, obs_delta: datetime.timedelta) -> Observations:

        start_time = time_now - obs_delta

        observations = self.client.fetch_observations(location=location, start=start_time, end=time_now)

        return observations

    def get_wd_forecast_from_now(self, location: Location) -> Forecast:

        forecast = self.client.fetch_forecast(location)

        return forecast

    def get_wd_now(self, location: Location, obs_delta: datetime.timedelta) -> WeatherData:

        time_now = datetime.datetime.now()

        observations = self.get_wd_observations_to_now(location, time_now, obs_delta)

        # print(observations)

        forecast = self.get_wd_forecast_from_now(location)

        # print(forecast)

        wd = WeatherData(created=time_now, observations=observations, forecast=forecast)

        return wd

    def compute_now(self, location: Location, obs_delta: datetime.timedelta) -> FireRiskPrediction:

        wd = self.get_wd_now(location, obs_delta)

        # print(wd.to_json())

        prediction = self.compute(wd)

        return prediction

    def compute_now_period(self, location: Location, obs_delta: datetime.timedelta, fct_delta: datetime.timedelta):
        pass

    def compute_period(self, location: Location, start: datetime, end: datetime) -> FireRiskPrediction:
        pass

    def compute_period_delta(self, location: Location, start: datetime, delta: datetime.timedelta) -> FireRiskPrediction:
        pass


class METFireRiskAPI:

    def __init__(self):
        self.met_extractor = METExtractor()

        self.met_client = METClient(extractor=self.met_extractor)

        self.frc = FireRiskAPI(client=self.met_client)

    def get_weatherdata_now(self, location: Location, obs_delta: datetime.timedelta) -> WeatherData:

        wd = self.frc.get_wd_now(location, obs_delta)

        return wd

    def compute(self, wd: WeatherData) -> FireRiskPrediction:
        return self.frc.compute(wd)

    def compute_now(self, location: Location, obs_delta: datetime.timedelta) -> FireRiskPrediction:
        return self.frc.compute_now(location, obs_delta)




import requests
import datetime
import json

# see .env.example.py in the root dir.
from decouple import config

from frcm.weatherdata.client import WeatherDataClient
from frcm.weatherdata.extractor import Extractor
from frcm.datamodel.model import Location, Observations, Forecast


class METClient(WeatherDataClient):

    def __init__(self, extractor: Extractor):

        self.forecast_endpoint = 'https://api.met.no/weatherapi/locationforecast/2.0/compact.json'

        self.observations_endpoint = 'https://frost.met.no/observations/v0.jsonld'
        self.sources_endpoint = 'https://frost.met.no/sources/v0.jsonld'

        self.MET_CLIENT_ID = config('MET_CLIENT_ID')
        self.MET_CLIENT_SECRET = config('MET_CLIENT_SECRET')

        self.extractor = extractor

    def send_met_request(self, parameters):

        header = {'User-Agent': 'DYNAMIC Firerisk Model'}

        response = requests.get(self.forecast_endpoint,
                                headers=header,
                                params=parameters,
                                auth=(self.MET_CLIENT_ID, self.MET_CLIENT_SECRET))

        return response

    def fetch_forecast_raw(self, location: Location):

        parameters = {'lat': str(location.latitude),
                      'lon': str(location.longitude)
                      }

        response = self.send_met_request(parameters)

        return response

    def fetch_forecast(self, location: Location) -> Forecast:

        response = self.fetch_forecast_raw(location)

        forecast = self.extractor.extract_forecast(response.text)

        return forecast

    def send_frost_request(self, endpoint, parameters):

        response = requests.get(endpoint,
                                params=parameters,
                                auth=(self.MET_CLIENT_ID, self.MET_CLIENT_SECRET))

        return response

    def get_nearest_station_raw(self, location: Location):
        parameters = {
            'types': 'SensorSystem',
            'elements': 'air_temperature,relative_humidity,wind_speed',
            'geometry':  f'nearest(POINT({location.longitude} {location.latitude}))'}

        response = self.send_frost_request(self.sources_endpoint, parameters)

        return response

    def get_nearest_station_id(self, location: Location) -> str:

        # TODO: more error handling here

        frost_response = self.get_nearest_station_raw(location)

        frost_response_str = frost_response.text

        station_response = json.loads(frost_response_str)

        station_id = station_response['data'][0]['id']

        return station_id

    @staticmethod
    def format_date(dt: datetime.datetime):

        return dt.strftime('%Y-%m-%d')

    @staticmethod
    def format_period(start: datetime.datetime, end: datetime.datetime):

        start_date = METClient.format_date(start)

        end_date = METClient.format_date(end)

        timeperiod = f'{start_date}/{end_date}'

        return timeperiod

    def fetch_observations_raw(self, source: str, start: datetime.datetime, end: datetime.datetime):

        time_period = METClient.format_period(start, end)

        print(f'Fetch observation : {time_period}')

        parameters = {'sources': source,
                      'referencetime': time_period,
                      'elements': 'air_temperature,relative_humidity,wind_speed'
                      }

        response = self.send_frost_request(self.observations_endpoint, parameters)

        return response

    def fetch_observations(self, location: Location, start: datetime.datetime, end: datetime.datetime) -> Observations:

#        print(location)

        station_id = self.get_nearest_station_id(location)

#        print(station_id)

        response = self.fetch_observations_raw(station_id, start, end)

#        print(response.text)

        observations = self.extractor.extract_observations(response.text, location)

        return observations
